@@ -28,6 +28,7 @@ module Data.Streaming.Network
     , HasPort (..)
     , HasAfterBind (..)
     , HasReadWrite (..)
+    , HasReadBufferSize (..)
 #if !WINDOWS
     , HasPath (..)
 #endif
@@ -37,6 +38,7 @@ module Data.Streaming.Network
     , setAddrFamily
     , setAfterBind
     , setNeedLocalAddr
+    , setReadBufferSize
 #if !WINDOWS
     , setPath
 #endif
@@ -46,6 +48,7 @@ module Data.Streaming.Network
     , getAddrFamily
     , getAfterBind
     , getNeedLocalAddr
+    , getReadBufferSize
 #if !WINDOWS
     , getPath
 #endif
@@ -53,6 +56,8 @@ module Data.Streaming.Network
     , appWrite
     , appSockAddr
     , appLocalAddr
+    , appCloseConnection
+    , appRawSocket
       -- * Functions
       -- ** General
     , bindPortGen
@@ -70,7 +75,7 @@ module Data.Streaming.Network
     , safeRecv
     , runTCPServer
     , runTCPClient
-    , ConnectionHandle (..)
+    , ConnectionHandle()
     , runTCPServerWithHandle
       -- ** UDP
     , bindPortUDP
@@ -101,9 +106,10 @@ import Data.Functor.Identity (Identity (Identity), runIdentity)
 import Control.Concurrent (forkIO)
 import Control.Monad (forever)
 import Data.IORef (IORef, newIORef, atomicModifyIORef)
-import Data.Array.Unboxed ((!), UArray, bounds, listArray)
+import Data.Array.Unboxed ((!), UArray, listArray)
 import System.IO.Unsafe (unsafePerformIO)
 import System.Random (randomRIO)
+import System.IO.Error (isFullErrorType, ioeGetErrorType)
 #if WINDOWS
 import Control.Concurrent.MVar (putMVar, takeMVar, newEmptyMVar)
 #endif
@@ -187,7 +193,7 @@ bindPortGen sockettype p s = do
 -- Since 0.1.1
 bindRandomPortGen :: SocketType -> HostPreference -> IO (Int, Socket)
 bindRandomPortGen sockettype s =
-    loop 30
+    loop (30 :: Int)
   where
     loop cnt = do
         port <- getUnassignedPort
@@ -269,6 +275,9 @@ bindPortUDP = bindPortGen NS.Datagram
 bindRandomPortUDP :: HostPreference -> IO (Int, Socket)
 bindRandomPortUDP = bindRandomPortGen NS.Datagram
 
+defaultReadBufferSize :: Int
+defaultReadBufferSize = 32768
+
 #if !WINDOWS
 -- | Attempt to connect to the given Unix domain socket path.
 getSocketUnix :: FilePath -> IO Socket
@@ -310,6 +319,7 @@ serverSettingsUnix
 serverSettingsUnix path = ServerSettingsUnix
     { serverPath = path
     , serverAfterBindUnix = const $ return ()
+    , serverReadBufferSizeUnix = defaultReadBufferSize
     }
 
 -- | Smart constructor.
@@ -318,6 +328,7 @@ clientSettingsUnix
     -> ClientSettingsUnix
 clientSettingsUnix path = ClientSettingsUnix
     { clientPath = path
+    , clientReadBufferSizeUnix = defaultReadBufferSize
     }
 #endif
 
@@ -356,6 +367,7 @@ serverSettingsTCP port host = ServerSettings
     , serverSocket = Nothing
     , serverAfterBind = const $ return ()
     , serverNeedLocalAddr = False
+    , serverReadBufferSize = defaultReadBufferSize
     }
 
 -- | Create a server settings that uses an already available listening socket.
@@ -369,6 +381,7 @@ serverSettingsTCPSocket lsocket = ServerSettings
     , serverSocket = Just lsocket
     , serverAfterBind = const $ return ()
     , serverNeedLocalAddr = False
+    , serverReadBufferSize = defaultReadBufferSize
     }
 
 -- | Smart constructor.
@@ -387,6 +400,7 @@ clientSettingsTCP port host = ClientSettings
     { clientPort = port
     , clientHost = host
     , clientAddrFamily = NS.AF_UNSPEC
+    , clientReadBufferSize = defaultReadBufferSize
     }
 
 -- | Attempt to connect to the given host/port/address family.
@@ -395,6 +409,7 @@ clientSettingsTCP port host = ClientSettings
 getSocketFamilyTCP :: ByteString -> Int -> NS.Family -> IO (NS.Socket, NS.SockAddr)
 getSocketFamilyTCP host' port' addrFamily = do
     (sock, addr) <- getSocketFamilyGen NS.Stream (S8.unpack host') port' addrFamily
+    NS.setSocketOption sock NS.NoDelay 1
     ee <- try' $ NS.connect sock (NS.addrAddress addr)
     case ee of
         Left e -> NS.sClose sock >> throwIO e
@@ -445,9 +460,12 @@ acceptSafe socket =
 #endif
   where
     loop =
-        NS.accept socket `E.catch` \(_ :: IOException) -> do
-            threadDelay 1000000
-            loop
+        NS.accept socket `E.catch` \e ->
+            if isFullErrorType (ioeGetErrorType e)
+                then do
+                    threadDelay 1000000
+                    loop
+                else E.throwIO e
 
 message :: ByteString -> NS.SockAddr -> Message
 message = Message
@@ -465,12 +483,6 @@ getPort = getConstant . portLens Constant
 setPort :: HasPort a => Int -> a -> a
 setPort p = runIdentity . portLens (const (Identity p))
 
-setHostPref :: HostPreference -> ServerSettings -> ServerSettings
-setHostPref hp ss = ss { serverHost = hp }
-
-getHostPref :: ServerSettings -> HostPreference
-getHostPref = serverHost
-
 setHost :: ByteString -> ClientSettings -> ClientSettings
 setHost hp ss = ss { clientHost = hp }
 
@@ -525,10 +537,40 @@ getAfterBind = getConstant . afterBindLens Constant
 setAfterBind :: HasAfterBind a => (Socket -> IO ()) -> a -> a
 setAfterBind p = runIdentity . afterBindLens (const (Identity p))
 
+-- | Since 0.1.13
+class HasReadBufferSize a where
+    readBufferSizeLens :: Functor f => (Int -> f Int) -> a -> f a
+-- | Since 0.1.13
+instance HasReadBufferSize ServerSettings where
+    readBufferSizeLens f ss = fmap (\p -> ss { serverReadBufferSize = p }) (f (serverReadBufferSize ss))
+-- | Since 0.1.13
+instance HasReadBufferSize ClientSettings where
+    readBufferSizeLens f cs = fmap (\p -> cs { clientReadBufferSize = p }) (f (clientReadBufferSize cs))
+#if !WINDOWS
+-- | Since 0.1.13
+instance HasReadBufferSize ServerSettingsUnix where
+    readBufferSizeLens f ss = fmap (\p -> ss { serverReadBufferSizeUnix = p }) (f (serverReadBufferSizeUnix ss))
+-- | Since 0.1.14
+instance HasReadBufferSize ClientSettingsUnix where
+    readBufferSizeLens f ss = fmap (\p -> ss { clientReadBufferSizeUnix = p }) (f (clientReadBufferSizeUnix ss))
+#endif
+
+-- | Get buffer size used when reading from socket.
+--
+-- Since 0.1.13
+getReadBufferSize :: HasReadBufferSize a => a -> Int
+getReadBufferSize = getConstant . readBufferSizeLens Constant
+
+-- | Set buffer size used when reading from socket.
+--
+-- Since 0.1.13
+setReadBufferSize :: HasReadBufferSize a => Int -> a -> a
+setReadBufferSize p = runIdentity . readBufferSizeLens (const (Identity p))
+
 type ConnectionHandle = Socket -> NS.SockAddr -> Maybe NS.SockAddr -> IO ()
 
-runTCPServerWithHandle :: ServerSettings -> ConnectionHandle -> IO ()
-runTCPServerWithHandle (ServerSettings port host msocket afterBind needLocalAddr) handle =
+runTCPServerWithHandle :: ServerSettings -> ConnectionHandle -> IO a
+runTCPServerWithHandle (ServerSettings port host msocket afterBind needLocalAddr _) handle =
     case msocket of
         Nothing -> E.bracket (bindPortTCP port host) NS.sClose inner
         Just lsocket -> inner lsocket
@@ -551,28 +593,32 @@ runTCPServerWithHandle (ServerSettings port host msocket afterBind needLocalAddr
 -- | Run an @Application@ with the given settings. This function will create a
 -- new listening socket, accept connections on it, and spawn a new thread for
 -- each connection.
-runTCPServer :: ServerSettings -> (AppData -> IO ()) -> IO ()
+runTCPServer :: ServerSettings -> (AppData -> IO ()) -> IO a
 runTCPServer settings app = runTCPServerWithHandle settings app'
   where app' socket addr mlocal =
           let ad = AppData
-                { appRead' = safeRecv socket 4096
+                { appRead' = safeRecv socket $ getReadBufferSize settings
                 , appWrite' = sendAll socket
                 , appSockAddr' = addr
                 , appLocalAddr' = mlocal
+                , appCloseConnection' = NS.sClose socket
+                , appRawSocket' = Just socket
                 }
           in
             app ad
 
 -- | Run an @Application@ by connecting to the specified server.
 runTCPClient :: ClientSettings -> (AppData -> IO a) -> IO a
-runTCPClient (ClientSettings port host addrFamily) app = E.bracket
+runTCPClient (ClientSettings port host addrFamily readBufferSize) app = E.bracket
     (getSocketFamilyTCP host port addrFamily)
     (NS.sClose . fst)
     (\(s, address) -> app AppData
-        { appRead' = safeRecv s 4096
+        { appRead' = safeRecv s readBufferSize
         , appWrite' = sendAll s
         , appSockAddr' = address
         , appLocalAddr' = Nothing
+        , appCloseConnection' = NS.sClose s
+        , appRawSocket' = Just s
         })
 
 appLocalAddr :: AppData -> Maybe NS.SockAddr
@@ -581,6 +627,19 @@ appLocalAddr = appLocalAddr'
 appSockAddr :: AppData -> NS.SockAddr
 appSockAddr = appSockAddr'
 
+-- | Close the underlying connection. One possible use case is simulating
+-- connection failures in a test suite.
+--
+-- Since 0.1.6
+appCloseConnection :: AppData -> IO ()
+appCloseConnection = appCloseConnection'
+
+-- | Get the raw socket for this @AppData@, if available.
+--
+-- Since 0.1.12
+appRawSocket :: AppData -> Maybe NS.Socket
+appRawSocket = appRawSocket'
+
 class HasReadWrite a where
     readLens :: Functor f => (IO ByteString -> f (IO ByteString)) -> a -> f a
     writeLens :: Functor f => ((ByteString -> IO ()) -> f (ByteString -> IO ())) -> a -> f a
@@ -603,8 +662,8 @@ appWrite = getConstant . writeLens Constant
 -- | Run an @Application@ with the given settings. This function will create a
 -- new listening socket, accept connections on it, and spawn a new thread for
 -- each connection.
-runUnixServer :: ServerSettingsUnix -> (AppDataUnix -> IO ()) -> IO ()
-runUnixServer (ServerSettingsUnix path afterBind) app = E.bracket
+runUnixServer :: ServerSettingsUnix -> (AppDataUnix -> IO ()) -> IO a
+runUnixServer (ServerSettingsUnix path afterBind readBufferSize) app = E.bracket
     (bindPath path)
     NS.sClose
     (\socket -> do
@@ -616,7 +675,7 @@ runUnixServer (ServerSettingsUnix path afterBind) app = E.bracket
         (\(socket, _) -> NS.sClose socket)
         $ \(socket, _) -> do
             let ad = AppDataUnix
-                    { appReadUnix = safeRecv socket 4096
+                    { appReadUnix = safeRecv socket readBufferSize
                     , appWriteUnix = sendAll socket
                     }
             _ <- E.mask $ \restore -> forkIO
@@ -626,11 +685,11 @@ runUnixServer (ServerSettingsUnix path afterBind) app = E.bracket
 
 -- | Run an @Application@ by connecting to the specified server.
 runUnixClient :: ClientSettingsUnix -> (AppDataUnix -> IO a) -> IO a
-runUnixClient (ClientSettingsUnix path) app = E.bracket
+runUnixClient (ClientSettingsUnix path readBufferSize) app = E.bracket
     (getSocketUnix path)
     NS.sClose
     (\sock -> app AppDataUnix
-        { appReadUnix = safeRecv sock 4096
+        { appReadUnix = safeRecv sock readBufferSize
         , appWriteUnix = sendAll sock
         })
 #endif
